# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import h5py
import numpy as np


class NwbDataSet(object):
    """ A very simple interface for exracting electrophysiology data
    from an NWB file.
    """
    SPIKE_TIMES = "spike_times"
    DEPRECATED_SPIKE_TIMES = "aibs_spike_times"

    def __init__(self, file_name, spike_time_key=None):
        """ Initialize the NwbDataSet instance with a file name

        Parameters
        ----------
        file_name: string
           NWB file name
        """
        self.file_name = file_name
        if spike_time_key is None:
            self.spike_time_key = NwbDataSet.SPIKE_TIMES
        else:
            self.spike_time_key = spike_time_key

    def get_sweep(self, sweep_number):
        """ Retrieve the stimulus, response, index_range, and sampling rate
        for a particular sweep.  This method hides the NWB file's distinction
        between a "Sweep" and an "Experiment".  An experiment is a subset of
        of a sweep that excludes the initial test pulse.  It also excludes
        any erroneous response data at the end of the sweep (usually for
        ramp sweeps, where recording was terminated mid-stimulus).

        Some sweeps do not have an experiment, so full data arrays are
        returned.  Sweeps that have an experiment return full data arrays
        (include the test pulse) with any erroneous data trimmed from the
        back of the sweep.

        Parameters
        ----------
        sweep_number: int

        Returns
        -------
        dict
            A dictionary with 'stimulus', 'response', 'index_range', and
            'sampling_rate' elements.  The index range is a 2-tuple where
            the first element indicates the end of the test pulse and the
            second index is the end of valid response data.
        """
        with h5py.File(self.file_name, 'r') as f:

            swp = f['epochs']['Sweep_%d' % sweep_number]

            # fetch data from file and convert to correct SI unit
            # this operation depends on file version. early versions of
            #   the file have incorrect conversion information embedded
            #   in the nwb file and data was stored in the appropriate
            #   SI unit. For those files, return uncorrected data.
            #   For newer files (1.1 and later), apply conversion value.
            major, minor = self.get_pipeline_version()
            if (major == 1 and minor > 0) or major > 1:
                # stimulus
                stimulus_dataset = swp['stimulus']['timeseries']['data']
                conversion = float(stimulus_dataset.attrs["conversion"])
                stimulus = stimulus_dataset.value * conversion
                # acquisition
                response_dataset = swp['response']['timeseries']['data']
                conversion = float(response_dataset.attrs["conversion"])
                response = response_dataset.value * conversion
            else:   # old file version
                stimulus_dataset = swp['stimulus']['timeseries']['data']
                stimulus = stimulus_dataset.value
                response = swp['response']['timeseries']['data'].value

            if 'unit' in stimulus_dataset.attrs:
                unit = stimulus_dataset.attrs["unit"]
                unit_str = None
                if unit.startswith('A'):
                    unit_str = "Amps"
                elif unit.startswith('V'):
                    unit_str = "Volts"
                assert unit_str is not None, Exception(
                    "Stimulus time series unit not recognized")
            else:
                unit = None
                unit_str = 'Unknown'

            swp_idx_start = swp['stimulus']['idx_start'].value
            swp_length = swp['stimulus']['count'].value

            swp_idx_stop = swp_idx_start + swp_length - 1
            sweep_index_range = (swp_idx_start, swp_idx_stop)

            # if the sweep has an experiment, extract the experiment's index
            # range
            try:
                exp = f['epochs']['Experiment_%d' % sweep_number]
                exp_idx_start = exp['stimulus']['idx_start'].value
                exp_length = exp['stimulus']['count'].value
                exp_idx_stop = exp_idx_start + exp_length - 1
                experiment_index_range = (exp_idx_start, exp_idx_stop)
            except KeyError:
                # this sweep has no experiment.  return the index range of the
                # entire sweep.
                experiment_index_range = sweep_index_range

            assert sweep_index_range[0] == 0, Exception(
                "index range of the full sweep does not start at 0.")

            return {
                'stimulus': stimulus,
                'response': response,
                'stimulus_unit' : unit_str,
                'index_range': experiment_index_range,
                'sampling_rate': 1.0 * swp['stimulus']['timeseries']['starting_time'].attrs['rate']
            }

    def set_sweep(self, sweep_number, stimulus, response):
        """ Overwrite the stimulus or response of an NWB file.
        If the supplied arrays are shorter than stored arrays,
        they are padded with zeros to match the original data
        size.

        Parameters
        ----------
        sweep_number: int

        stimulus: np.array
           Overwrite the stimulus with this array.  If None, stimulus is unchanged.

        response: np.array
            Overwrite the response with this array.  If None, response is unchanged.
        """

        with h5py.File(self.file_name, 'r+') as f:
            swp = f['epochs']['Sweep_%d' % sweep_number]

            # this is the length of the entire sweep data, including test pulse and
            # whatever might be in front of it
            # TODO: remove deprecated 'idx_stop'
            if 'idx_stop' in swp['stimulus']:
                sweep_length = swp['stimulus']['idx_stop'].value + 1
            else:
                sweep_length = swp['stimulus']['count'].value

            if stimulus is not None:
                # if the data is shorter than the sweep, pad it with zeros
                missing_data = sweep_length - len(stimulus)
                if missing_data > 0:
                    stimulus = np.append(stimulus, np.zeros(missing_data))

                swp['stimulus']['timeseries']['data'][...] = stimulus

            if response is not None:
                # if the data is shorter than the sweep, pad it with zeros
                missing_data = sweep_length - len(response)
                if missing_data > 0:
                    response = np.append(response, np.zeros(missing_data))

                swp['response']['timeseries']['data'][...] = response

    def get_pipeline_version(self):
        """ Returns the AI pipeline version number, stored in the 
            metadata field 'generated_by'. If that field is
            missing, version 0.0 is returned.

            Returns
            -------
            int tuple: (major, minor)
        """
        try:
            with h5py.File(self.file_name, 'r') as f:
                if 'generated_by' in f["general"]:
                    info = f["general/generated_by"]
                    # generated_by stores array of keys and values
                    # keys are even numbered, corresponding values are in
                    #   odd indices
                    for i in range(len(info)):
                        val = info[i]
                        if info[i] == 'version':
                            version = info[i+1]
                            break
            toks = version.split('.')
            if len(toks) >= 2:
                major = int(toks[0])
                minor = int(toks[1])
        except:
            minor = 0
            major = 0
        return major, minor

    def get_spike_times(self, sweep_number, key=None):
        """ Return any spike times stored in the NWB file for a sweep.

        Parameters
        ----------
        sweep_number: int
            index to access
        key : string
            label where the spike times are stored (default NwbDataSet.SPIKE_TIMES)

        Returns
        -------
        list
           list of spike times in seconds relative to the start of the sweep
        """

        if key is None:
            key = self.spike_time_key

        with h5py.File(self.file_name, 'r') as f:
            sweep_name = "Sweep_%d" % sweep_number
            datasets = ["analysis/%s/Sweep_%d" % (key, sweep_number),
                        "analysis/%s/Sweep_%d" % (self.DEPRECATED_SPIKE_TIMES, sweep_number)]

            for ds in datasets:
                if ds in f:
                    return f[ds].value
            return []

    def set_spike_times(self, sweep_number, spike_times, key=None):
        """ Set or overwrite the spikes times for a sweep.

        Parameters
        ----------
        sweep_number : int
            index to access
        key : string
            where the times are stored (default NwbDataSet.SPIKE_TIME)

        spike_times: np.array
           array of spike times in seconds
        """

        if key is None:
            key = self.spike_time_key

        with h5py.File(self.file_name, 'r+') as f:
            # make sure expected directory structure is in place
            if "analysis" not in f.keys():
                f.create_group("analysis")

            analysis_dir = f["analysis"]
            if NwbDataSet.SPIKE_TIMES not in analysis_dir.keys():
                #   analysis_dir.create_group(NwbDataSet.SPIKE_TIMES)
                g = analysis_dir.create_group(NwbDataSet.SPIKE_TIMES)
                # mixup in specification for validator resulted everything
                #   in 'analysis' requiring a custom label, even though
                #   it's already known to be custom. don't argue, just
                #   support the metadata redundancy
                g.attrs["neurodata_type"] = "Custom"

            spike_dir = analysis_dir[NwbDataSet.SPIKE_TIMES]

            # see if desired dataset already exists
            sweep_name = "Sweep_%d" % sweep_number
            if sweep_name in spike_dir.keys():
                # rewriting data -- delete old dataset
                del spike_dir[sweep_name]

            spike_dir.create_dataset(
                sweep_name, data=spike_times, dtype='f8', maxshape=(None,))

    def get_sweep_numbers(self):
        """ Get all of the sweep numbers in the file, including test sweeps. """

        with h5py.File(self.file_name, 'r') as f:
            sweeps = [int(e.split('_')[1])
                      for e in f['epochs'].keys() if e.startswith('Sweep_')]
            return sweeps

    def get_experiment_sweep_numbers(self):
        """ Get all of the sweep numbers for experiment epochs in the file, not including test sweeps. """

        with h5py.File(self.file_name, 'r') as f:
            sweeps = [int(e.split('_')[1])
                      for e in f['epochs'].keys() if e.startswith('Experiment_')]
            return sweeps

    def fill_sweep_responses(self, fill_value=0.0, sweep_numbers=None):
        """ Fill sweep response arrays with a single value.

        Parameters
        ----------
        fill_value: float
            Value used to fill sweep response array

        sweep_numbers: list
            List of integer sweep numbers to be filled (default all sweeps)

        """

        with h5py.File(self.file_name, 'a') as f:
            if sweep_numbers is None:
                # no sweep numbers given, grab all of them
                epochs = [k for k in f[
                    'epochs'].keys() if k.startswith('Sweep_')]
            else:
                epochs = ['Sweep_%d' %
                          sweep_number for sweep_number in sweep_numbers]

            for epoch in epochs:
                if epoch in f['epochs']:
                    f['epochs'][epoch]['response'][
                        'timeseries']['data'][...] = fill_value

    def get_sweep_metadata(self, sweep_number):
        """ Retrieve the sweep level metadata associated with each sweep.
        Includes information on stimulus parameters like its name and amplitude
        as well as recording quality metadata, like access resistance and
        seal quality.

        Parameters
        ----------
        sweep_number: int

        Returns
        -------
        dict
            A dictionary with 'aibs_stimulus_amplitude_pa', 'aibs_stimulus_name',
            'gain', 'initial_access_resistance', 'seal' elements.  These specific
            fields are ones encoded in the original AIBS in vitro .nwb files.
        """
        with h5py.File(self.file_name, 'r') as f:

            sweep_metadata = {}

            # the sweep level metadata is stored in
            # stimulus/presentation/Sweep_XX in the .nwb file

            # indicates which metadata fields to return
            metadata_fields = ['aibs_stimulus_amplitude_pa', 'aibs_stimulus_name',
                               'gain', 'initial_access_resistance', 'seal']
            try:
                stim_details = f['stimulus']['presentation'][
                    'Sweep_%d' % sweep_number]
                for field in metadata_fields:
                    # check if sweep contains the specific metadata field
                    if field in stim_details.keys():
                        sweep_metadata[field] = stim_details[field].value

            except KeyError:
                sweep_metadata = {}

            return sweep_metadata



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from . import runner as single_cell
import logging
import os
import sys
import traceback
import subprocess
import logging.config as lc
from ..biophys_sim.config import Config
from pkg_resources import resource_filename  # @UnresolvedImport


class RunSimulate(object):
    _log = logging.getLogger('allensdk.model.biophysical.run_simulate')

    def __init__(self,
                 input_json,
                 output_json):
        self.input_json = input_json
        self.output_json = output_json
        self.app_config = None
        self.manifest = None

    def load_manifest(self):
        self.app_config = Config().load(self.input_json)
        self.manifest = self.app_config.manifest
        fix_sections = ['passive', 'axon_morph,', 'conditions', 'fitting']
        self.app_config.fix_unary_sections(fix_sections)

    def nrnivmodl(self):
        RunSimulate._log.debug("nrnivmodl")

        subprocess.call(['nrnivmodl', './modfiles'])

    def simulate(self):
        from allensdk.internal.api.queries.biophysical_module_reader \
            import BiophysicalModuleReader

        self.load_manifest()

        try:
            stimulus_path = self.manifest.get_path('stimulus_path')
            RunSimulate._log.info("stimulus path: %s" % (stimulus_path))
        except:
            raise Exception(
                'Could not read input stimulus path from input config.')

        try:
            out_path = self.manifest.get_path('output_path')
            RunSimulate._log.info("result NWB file: %s" % (out_path))
        except:
            raise Exception('Could not read output path from input config.')

        try:
            morphology_path = self.manifest.get_path('MORPHOLOGY')
            RunSimulate._log.info("morphology path: %s" % (morphology_path))
        except:
            raise Exception(
                'Could not read morphology path from input config.')

        single_cell.run(self.app_config)

        lims_upload_config = BiophysicalModuleReader()
        lims_upload_config.read_json(
            self.manifest.get_path('neuronal_model_run_data'))
        lims_upload_config.update_well_known_file(out_path)
        lims_upload_config.set_workflow_state('passed')
        lims_upload_config.write_file(self.output_json)


def main(command, lims_strategy_json, lims_response_json):
    ''' Entry point for module.
        :param command: select behavior, nrnivmodl or simulate
        :type command: string
        :param lims_strategy_json: path to json file output from lims.
        :type lims_strategy_json: string
        :param lims_response_json: path to json file returned to lims.
        :type lims_response_json: string
    '''
    rs = RunSimulate(lims_strategy_json,
                     lims_response_json)

    RunSimulate._log.debug("command: %s" % (command))
    RunSimulate._log.debug("lims strategy json: %s" % (lims_strategy_json))
    RunSimulate._log.debug("lims upload json: %s" % (lims_response_json))

    log_config = resource_filename('allensdk.model.biophysical.run_simulate',
                                   'logging.conf')
    lc.fileConfig(log_config)
    os.environ['LOG_CFG'] = log_config

    if 'nrnivmodl' == command:
        rs.nrnivmodl()
    else:
        rs.simulate()


if __name__ == '__main__':
    command, input_json, output_json = sys.argv[-3:]

    try:
        main(command, input_json, output_json)
        RunSimulate._log.debug("success")
    except Exception as e:
        RunSimulate._log.error(traceback.format_exc())
        exit(1)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
####test AllenSDK tree search api for Specimen and Structure
from allensdk.api.queries.tree_search_api import TreeSearchApi
import pytest
import json
from mock import MagicMock

@pytest.fixture
def tree_search():
    tsa = TreeSearchApi()
    tsa.json_msg_query = MagicMock(name='json_msg_query')

    return tsa

def test_get_specimen_tree(tree_search):
    ####ancestor true for Specimen
    kind = 'Specimen'
    db_id = 113817886
    ancestors = True
    descendants = None
    tree_search.get_tree(kind, db_id, ancestors, descendants)
    tree_search.json_msg_query.assert_called_with("http://api.brain-map.org/api/v2/tree_search/Specimen/113817886.json?ancestors=true")
    
    ####ancestor true for Specimen
    kind = 'Specimen'
    db_id = 113817886
    ancestors = True
    descendants = False
    tree_search.get_tree(kind, db_id, ancestors, descendants)
    tree_search.json_msg_query.assert_called_with("http://api.brain-map.org/api/v2/tree_search/Specimen/113817886.json?ancestors=true&descendants=false")
    
    ####ancestor false for Specimen
    kind = 'Specimen'
    db_id = 113817886
    ancestors = False
    descendants = True
    tree_search.get_tree(kind, db_id, ancestors, descendants)
    tree_search.json_msg_query.assert_called_with("http://api.brain-map.org/api/v2/tree_search/Specimen/113817886.json?ancestors=false&descendants=true")

def test_get_structure_tree(tree_search):
    ####ancestor True for Structure
    kind = 'Structure'
    db_id = 12547
    ancestors = True
    descendants = True
    tree_search.get_tree(kind, db_id, ancestors, descendants)
    tree_search.json_msg_query.assert_called_with("http://api.brain-map.org/api/v2/tree_search/Structure/12547.json?ancestors=true&descendants=true")
    
    ####ancestor False for Structure
    kind = 'Structure'
    db_id = 12547
    ancestors = False
    descendants = True
    tree_search.get_tree(kind, db_id, ancestors, descendants)
    tree_search.json_msg_query.assert_called_with("http://api.brain-map.org/api/v2/tree_search/Structure/12547.json?ancestors=false&descendants=true")
    
    ####ancestor None for Structure
    kind = 'Structure'
    db_id = 12547
    ancestors = None
    descendants = None
    tree_search.get_tree(kind, db_id, ancestors, descendants)
    tree_search.json_msg_query.assert_called_with("http://api.brain-map.org/api/v2/tree_search/Structure/12547.json")



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2014-2015. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from allensdk.core.json_utilities import JsonComments
import argparse
import os
import io
import logging
import logging.config as lc
from pkg_resources import resource_filename  # @UnresolvedImport
try:
    from configparser import ConfigParser  # @UnresolvedImport
except:
    from ConfigParser import ConfigParser  # @UnresolvedImport


class ApplicationConfig(object):
    ''' Convenience class that handles of application configuration
    from environment variables, .conf files and the command line
    using Python standard libraries and formats.
    '''

    _log = logging.getLogger(__name__)
    _DEFAULT_LOG_CONFIG = os.getenv(
        'LOG_CFG', resource_filename(__name__, 'logging.conf'))
    lc.fileConfig(_DEFAULT_LOG_CONFIG)

    def __init__(self,
                 defaults,
                 name="app",
                 halp="Run application.",
                 default_log_config=None):
        self.application_name = name
        self.help = halp
        self.debug_enabled = False

        if default_log_config is None:
            default_log_config = ApplicationConfig._DEFAULT_LOG_CONFIG

        ApplicationConfig._log.info(
            "default log config: %s" % (default_log_config))

        self.defaults = {
            'config_file_path': {
                'default': "%s.conf" % (self.application_name),
                'help': 'configuration file path'
            },
            'log_config_path': {
                'default': default_log_config,
                'help': 'logging configuration path'
            }
        }

        self.defaults.update(defaults)

        logging.info("defaults: %s" % (self.defaults))

        self.argparser = self.create_argparser()

        for key, value in self.defaults.items():
            setattr(self, key, value['default'])

    def load(self, command_line_args, disable_existing_loggers=True):
        ''' Load application configuration options, first from the environment,
        then from the configuration file, then from the command line.

        Each stage of loading can override the previous stage.

        Parameters
        ----------
        command_line_args : dict
            Parameters passed to the application.
        disable_existing_loggers : boolean
            Reset the logging system or not.

        Returns
        -------
        fileConfig
            Configuration object with all levels applied
        '''
        # read and apply options from the environment
        self.apply_configuration_from_environment()

        # command line so we can find the config file.
        parsed_args = self.parse_command_line_args(command_line_args)

        try:
            # read and apply the configuration file options
            config_file_path = parsed_args.config_file_path

            if config_file_path:
                self.config_file_path = config_file_path

            self.apply_configuration_from_file(self.config_file_path)

            # apply the remaining command line options
            self.apply_configuration_from_command_line(parsed_args)
        except Exception as e:
            ApplicationConfig._log.error("Could not load configuration file: %s\n%s" %
                                         (parsed_args.config_file_path,
                                          e))
            raise

        if parsed_args.log_config_path:
            try:
                lc.fileConfig(self.log_config_path,
                              disable_existing_loggers=disable_existing_loggers)
            except:
                logging.error("Could not load log configuration file: %s" %
                              (parsed_args.log_config_path))
        else:
            # TODO: configure default logging
            pass

    def create_argparser(self):
        '''Initialization for the command-line parsing stage.

        An application specific prefix is applied to argument names.

        Parameters
        ----------
        prog : string
            Application specific prefix for argument names.
        description : string
            A brief 'help' description of the application.

        Returns
        -------
        argParse.ArgumentParser
            The initialized argument parser object.

        Notes
        -----
        Defaults are set at the first environment reading.
        Command line args only override them when present
        '''
        parser = argparse.ArgumentParser(prog=self.application_name,
                                         description=self.help)
        for key, value in self.defaults.items():
            if key == 'config_file_path':
                parser.add_argument(
                    "%s" % (key), default=None, help=value['help'])
            else:
                parser.add_argument("--%s" %
                                    (key), default=None, help=value['help'])

        return parser

    def parse_command_line_args(self, args):
        '''Simply call the internal argparser object.

        Parameters
        ----------
        args : array
            Parameters passed to the application.

        Returns
        -------
        Namespace
            Parsed paramenters.
        '''
        return self.argparser.parse_args(args)

    def apply_configuration_from_command_line(self, parsed_args):
        '''Read application configuration variables from the command line.

        Unassigned variables are left unchanged if previously assigned,
        set to their default values,
        or None if no default is specified at init time.
        Assigned variables will overwrite the previous value.

        see: https://docs.python.org/2/howto/argparse.html

        Parameters
        ----------
        parsed_args : dict
            the arguments as parsed from the command line.

        '''
        logging.info('command_line args: %s' % (parsed_args))

        for key in self.defaults:
            parsed_value = getattr(parsed_args, key)
            if parsed_value and getattr(self, key) is None:
                setattr(self, key, parsed_value)

    def apply_configuration_from_environment(self):
        '''Read application configuration variables from the environment.

        The variable names are upper case and have a
        prefix defined by the application.

        See: https://docs.python.org/2/library/os.html
        '''
        for key in self.defaults:
            environment_variable = "%s_%s" % (
                self.application_name.upper(), key.upper())
            environment_value = os.environ.get(environment_variable)
            if environment_value:
                setattr(self, key, environment_value)

    def from_json_file(self, json_path):
        '''Read an application configuration from a JSON format file.

        Parameters
        ----------
        json_path : string
            Path to the JSON file.

        Returns
        -------
        string
            An application configuration in INI format

        '''
        description = JsonComments.read_file(json_path)

        return self.to_config_string(description)

    def from_json_string(self, json_string):
        '''Read a configuration from a JSON format string.

        Parameters
        ----------
        json_string : string
            A JSON-formatted string containing an application configuration.

        Returns
        -------
        string
            An application configuration in INI format
        '''
        description = JsonComments.read_string(json_string)

        return self.to_config_string(description)

    def to_config_string(self, description):
        '''Create a configuration string from a dict.

        Parameters
        ----------
        description : dict
            Configuration options for an application.

        Returns
        -------
        string
            Equivalent configuration as an INI format string

        Notes
        -----
        The Python configparser library natively supports this functionality in Python 3.
        '''
        if 'biophys' not in description:
            bps_config_string = '[biophys]\n\n'
            return bps_config_string

        bps_config = description['biophys'][0]

        cfg_array = ['[biophys]']

        if 'log_config_path' in bps_config:
            cfg_array.append(str('log_config_path: %s' %
                                 bps_config['log_config_path']))

        if 'debug' in bps_config:
            cfg_array.append(str('debug: %s' % bps_config['debug']))

        if 'model_file' in bps_config:
            cfg_array.append(str('model_file: %s' %
                                 ','.join(bps_config['model_file'])))

        cfg_array.append("\n")

        bps_cfg_string = "\n".join(cfg_array)
        ApplicationConfig._log.info(bps_cfg_string)

        return bps_cfg_string

    def apply_configuration_from_file(self, config_file_path):
        ''' Read application configuration variables from a .conf file.

        Unassigned variables are set to their default values
        or None if no default is specified at init time.
        The variables are found in a section named by the application.

        Parameters
        ----------
        config_file_path : string
            path to to an INI (.conf) or JSON format application config file.

        Returns
        -------

        see: https://docs.python.org/2/library/configparser.html
        '''
        none_defaults = {}

        # defaults are set in environment
        # they are only overriden by the config file if present
        for key in self.defaults:
            none_defaults[key] = None

        logging.info("none_defaults: %s" % (none_defaults))

        config = None

        try:
            config = ConfigParser(defaults=none_defaults,
                                  allow_no_value=True)
        except:
            logging.warn(
                "This python installation does not support configuration defaults.")
            config = ConfigParser()

        if config_file_path.endswith('.json'):
            cfg_string = self.from_json_file(config_file_path)
            try:
                config.readfp(io.BytesIO(cfg_string))
            except (NameError, TypeError):
                config.read_string(cfg_string)  # Python 3
        else:
            config.read(config_file_path)

        for key in self.defaults:
            try:
                file_value = config.get(self.application_name, key)
                if file_value:
                    logging.info("setting %s to %s" % (key, file_value))
                    setattr(self, key, file_value)
            except:
                logging.info("Configuration option not specified: %s" %
                             (key))



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from .fit_parameters import get_gaussian_fit_single_channel, compute_distance, compute_overlap
from .chisquarerf import chi_square_binary, get_peak_significance, pvalue_to_NLL
from .utilities import upsample_image_to_degrees
import collections
import numpy as np
import sys

def get_gaussian_fit(rf):

    fit_parameters_dict_combined = {'on':collections.defaultdict(list), 'off':collections.defaultdict(list)}
    counter = {'on':0, 'off':0}
    for on_off_key in ['on', 'off']:
        fit_parameters_dict = fit_parameters_dict_combined[on_off_key]
        for ci in range(rf[on_off_key]['fdr_mask']['attrs']['number_of_components']):
            curr_component_mask = upsample_image_to_degrees(np.logical_not(rf[on_off_key]['fdr_mask']['data'][ci,:,:])) > .5
            rf_response = upsample_image_to_degrees(rf[on_off_key]['rts_convolution']['data'].copy())
            rf_response[curr_component_mask] = 0

            if rf_response.sum() > 0:
                get_gaussian_fit_single_channel(rf_response, fit_parameters_dict)
                counter[on_off_key] += 1

    for ii_off in range(counter['on']):
        fit_parameters_dict_combined['on']['distance'].append([None]*counter['off'])
        fit_parameters_dict_combined['on']['overlap'].append([None] * counter['off'])

    for ii_off in range(counter['off']):
        fit_parameters_dict_combined['off']['distance'].append([None] * counter['on'])
        fit_parameters_dict_combined['off']['overlap'].append([None]*counter['on'])


    for ii_on in range(counter['on']):
        for ii_off in range(counter['off']):
            center_on = fit_parameters_dict_combined['on']['center_x'][ii_on], fit_parameters_dict_combined['on']['center_y'][ii_on]
            center_off = fit_parameters_dict_combined['off']['center_x'][ii_off], fit_parameters_dict_combined['off']['center_y'][ii_off]
            curr_distance = compute_distance(center_on, center_off)
            fit_parameters_dict_combined['on']['distance'][ii_on][ii_off] = curr_distance
            fit_parameters_dict_combined['off']['distance'][ii_off][ii_on] = curr_distance

            data_on = fit_parameters_dict_combined['on']['data'][ii_on]
            data_off = fit_parameters_dict_combined['off']['data'][ii_off]
            curr_overlap = compute_overlap(data_on, data_off)
            fit_parameters_dict_combined['on']['overlap'][ii_on][ii_off] = curr_overlap
            fit_parameters_dict_combined['off']['overlap'][ii_off][ii_on] = curr_overlap

    return fit_parameters_dict_combined, counter

def run_postprocessing(data, rf):

    stimulus = rf['attrs']['stimulus']

    # Gaussian fit postprocessing:
    fit_parameters_dict_combined, counter = get_gaussian_fit(rf)

    for on_off_key in ['on', 'off']:

        if counter[on_off_key] > 0:

            rf[on_off_key]['gaussian_fit'] = {}
            rf[on_off_key]['gaussian_fit']['attrs'] = {}

            fit_parameters_dict = fit_parameters_dict_combined[on_off_key]
            for key, val in fit_parameters_dict.items():

                if key == 'data':
                    rf[on_off_key]['gaussian_fit']['data'] = np.array(val)
                else:
                    rf[on_off_key]['gaussian_fit']['attrs'][key] = np.array(val)

    # Chi squared test statistic postprocessing:
    cell_index = rf['attrs']['cell_index']
    locally_sparse_noise_template = data.get_stimulus_template(stimulus)

    event_array = np.zeros((rf['event_vector']['data'].shape[0], 1), dtype=np.bool)
    event_array[:,0] = rf['event_vector']['data']

    chi_squared_grid = chi_square_binary(event_array, locally_sparse_noise_template)
    alpha = rf['on']['fdr_mask']['attrs']['alpha']
    assert rf['off']['fdr_mask']['attrs']['alpha'] == alpha
    chi_square_grid_NLL = pvalue_to_NLL(chi_squared_grid)

    peak_significance = get_peak_significance(chi_square_grid_NLL, locally_sparse_noise_template,  alpha=alpha)
    significant = peak_significance[0][0]
    min_p = peak_significance[1][0]
    pvalues_chi_square = peak_significance[2][0]
    best_exclusion_region_mask = peak_significance[3][0]

    chi_squared_grid_dict = {
                             'best_exclusion_region_mask':{'data':best_exclusion_region_mask},
                             'attrs':{'significant':significant, 'alpha': alpha, 'min_p':min_p},
                             'pvalues':{'data':pvalues_chi_square}
                             }

    rf['chi_squared_analysis'] = chi_squared_grid_dict

    return rf

if __name__ == "__main__":
    # csid = 517472416 # triple!
    csid = 517526760 # two ON
    # csid = 539917553
    # csid = 540988186



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import MagicMock
from allensdk.api.queries.image_download_api import ImageDownloadApi


@pytest.fixture
def image_api():
    image_api = ImageDownloadApi()

    image_api.retrieve_file_over_http = \
        MagicMock(name='retrieve_file_over_http')
    image_api.json_msg_query = MagicMock(name='json_msg_query')

    return image_api


def test_api_doc_url_download_section_image_downsampled(image_api):
    '''
    Notes
    -----
    See: `Experimental Overview and Metadata `<http://help.brain-map.org/display/mouseconnectivity/API#API-ExperimentalOverviewandMetadata>_
    , link labeled 'Download image downsampled by factor of 6 using default thresholds'.
    '''
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=6,
                                     range=[0, 932, 0, 1279, 0, 4095])

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=6&range=0,932,0,1279,0,4095",
        path)


def test_api_doc_url_download_section_image_downsample_dimensions(image_api):
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=6,
                                     downsample_dimensions=True)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=6&downsample_dimensions=true",
        path)


def test_api_doc_url_download_section_image_full_res(image_api):
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575",
        path)


def test_api_doc_url_download_section_image_downsample_dimensions_false(image_api):
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=6,
                                     downsample_dimensions=False)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=6&downsample_dimensions=false",
        path)


def test_api_doc_url_download_section_image_downsampled_low_quality(image_api):
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=3,
                                     quality=50)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=3&quality=50",
        path)


def test_api_doc_url_download_section_image_tumor_feature_annotation(image_api):
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=6,
                                     tumor_feature_annotation=True)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=6&tumor_feature_annotation=true",
        path)


def test_api_doc_url_download_section_image_tumor_feature_annotation_false(image_api):
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=6,
                                     tumor_feature_annotation=False)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=6&tumor_feature_annotation=false",
        path)


def test_api_doc_url_download_section_image_tumor_feature_boundary(image_api):
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=6,
                                     tumor_feature_boundary=True)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=6&tumor_feature_boundary=true",
        path)


def test_api_doc_url_download_section_image_tumor_feature_boundary_false(image_api):
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=6,
                                     tumor_feature_boundary=False)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=6&tumor_feature_boundary=false",
        path)


def test_api_doc_url_download_section_image_expression(image_api):
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=6,
                                     expression=True)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=6&expression=true",
        path)


def test_api_doc_url_download_section_image_expression_false(image_api):
    path = '126862575.jpg'

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     downsample=6,
                                     expression=False)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/126862575"
        "?downsample=6&expression=false",
        path)


def test_api_doc_url_download_atlas_image_downsampled(image_api):
    path = '100883869.jpg'

    section_image_id = 100883869
    image_api.download_atlas_image(section_image_id,
                                   downsample=4)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/atlas_image_download/100883869"
        "?downsample=4",
        path)


def test_api_doc_url_download_atlas_image_downsampled_low_quality(image_api):
    path = '100883869.jpg'

    section_image_id = 100883869
    image_api.download_atlas_image(section_image_id,
                                   downsample=4,
                                   quality=50)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/atlas_image_download/100883869"
        "?downsample=4&quality=50",
        path)


def test_api_doc_url_download_atlas_image_annotation(image_api):
    path = '100883869.jpg'

    section_image_id = 100883869
    image_api.download_atlas_image(section_image_id,
                                   downsample=4,
                                   annotation=True)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/atlas_image_download/100883869"
        "?downsample=4&annotation=true",
        path)


def test_api_doc_url_download_atlas_image_annotation_false(image_api):
    path = '100883869.jpg'

    section_image_id = 100883869
    image_api.download_atlas_image(section_image_id,
                                   downsample=4,
                                   annotation=False)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/atlas_image_download/100883869"
        "?downsample=4&annotation=false",
        path)


def test_api_doc_url_download_atlas_image_atlas(image_api):
    path = '100883869.jpg'

    section_image_id = 100883869
    image_api.download_atlas_image(section_image_id,
                                   downsample=4,
                                   annotation=True,
                                   atlas=2)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/atlas_image_download/100883869"
        "?downsample=4&annotation=true&atlas=2",
        path)


def test_api_doc_url_download_atlas_full_resolution_region_of_interest(image_api):
    path = '100883869.jpg'

    subimage_id = 100883869
    image_api.download_atlas_image(subimage_id,
                                   left=6174,
                                   top=2282,
                                   width=1000,
                                   height=1000)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/atlas_image_download/100883869"
        "?left=6174&top=2282&width=1000&height=1000",
        path)


def test_api_doc_url_download_projection_image_downsampled(image_api):
    path = '126862583.jpg'

    section_image_id = 126862583
    image_api.download_projection_image(section_image_id,
                                        downsample=4)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/projection_image_download/126862583"
        "?downsample=4",
        path)


def test_api_doc_url_download_projection_image_projection(image_api):
    path = '126862583.jpg'

    section_image_id = 126862583
    image_api.download_projection_image(section_image_id,
                                        downsample=4,
                                        projection=True)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/projection_image_download/126862583"
        "?downsample=4&projection=true",
        path)


def test_api_doc_url_download_projection_image_projection_false(image_api):
    path = '126862583.jpg'

    section_image_id = 126862583
    image_api.download_projection_image(section_image_id,
                                        downsample=4,
                                        projection=False)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/projection_image_download/126862583"
        "?downsample=4&projection=false",
        path)


def test_api_doc_url_download_projection_image_view(image_api):
    path = '126862583.jpg'

    section_image_id = 126862583
    image_api.download_projection_image(section_image_id,
                                        downsample=4,
                                        view='projection')

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/projection_image_download/126862583"
        "?downsample=4&view=projection",
        path)


def test_api_doc_url_download_projection_image_view_exception(image_api):
    path = '126862583.jpg'

    section_image_id = 126862583
    
    with pytest.raises(ValueError) as excinfo:
        image_api.download_projection_image(section_image_id,
                                            downsample=4,
                                            view='typo')

    assert excinfo.value.args[0] == "view argument should be 'expression', 'projection', 'tumor_feature_annotation' or 'tumor_feature_boundary'"


def test_api_doc_url_download_image_downsampled(image_api):
    '''
    Notes
    -----
    See: `Image Download Service `<http://help.brain-map.org/display/api/Downloading+an+Image>_
    '''
    path = '69750516.jpg'

    subimage_id = 69750516
    image_api.download_image(subimage_id,
                             downsample=4)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/image_download/69750516"
        "?downsample=4",
        path)


def test_api_doc_url_download_image_downsampled_low_quality(image_api):
    '''
    Notes
    -----
    See: `Image Download Service `<http://help.brain-map.org/display/api/Downloading+an+Image>_
    '''
    path = '69750516.jpg'

    subimage_id = 69750516
    image_api.download_image(subimage_id,
                             downsample=3,
                             quality=50)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/image_download/69750516"
        "?downsample=3&quality=50",
        path)


def test_api_doc_url_download_full_resolution_region_of_interest(image_api):
    '''
    Notes
    -----
    See: `Image Download Service `<http://help.brain-map.org/display/api/Downloading+an+Image>_
    '''
    path = '69750516.jpg'

    subimage_id = 69750516
    image_api.download_image(subimage_id,
                             left=6174,
                             top=2282,
                             width=1000,
                             height=1000)

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/image_download/69750516"
        "?left=6174&top=2282&width=1000&height=1000",
        path)


def test_api_doc_url_download_image_expression_mask(image_api):
    '''
    Notes
    -----
    See: `Image Download Service `<http://help.brain-map.org/display/api/Downloading+an+Image>_
    '''
    path = '69750516.jpg'

    subimage_id = 69750516
    image_api.download_image(subimage_id,
                             downsample=4,
                             view='expression')

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/image_download/69750516"
        "?downsample=4&view=expression",
        path)


def test_api_doc_url_download_image_full_resolution(image_api):
    '''
    Notes
    -----
    See: `Experimental Overview and Metadata `<http://help.brain-map.org/display/mouseconnectivity/API#API-ExperimentalOverviewandMetadata>_
    , link labeled 'Download a region of interest at full resolution using default thresholds'.
    '''
    expected = 'http://api.brain-map.org/api/v2/section_image_download/126862575?range=0,932,0,1279,0,4095&left=19045&top=11684&width=1000&height=1000'
    path = '126862575.jpg'

    image_api.retrieve_file_over_http = \
        MagicMock(name='retrieve_file_over_http')

    section_image_id = 126862575
    image_api.download_section_image(section_image_id,
                                     left=19045,
                                     top=11684,
                                     width=1000,
                                     height=1000,
                                     range=[0, 932, 0, 1279, 0, 4095])

    image_api.retrieve_file_over_http.assert_called_once_with(expected, path)


def test_colormap_filter(image_api):
    '''
    '''
    path = '70636013.jpg'

    section_image_id = 70636013
    image_api.download_section_image(section_image_id,
                                     downsample=4,
                                     view='expression',
                                     colormap=(0.9,"expression"))

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/70636013"
        "?downsample=4&colormap=0.5,0.9,0,256,4&view=expression",
        path)


def test_colormap_filter_string(image_api):
    '''
    '''
    path = '70636013.jpg'

    section_image_id = 70636013
    image_api.download_section_image(section_image_id,
                                     downsample=4,
                                     view='expression',
                                     colormap="expression")

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/70636013"
        "?downsample=4&colormap=expression&view=expression",
        path)



def test_rgb_filter(image_api):
    '''
    '''
    path = '70636013.jpg'

    section_image_id = 70636013
    image_api.download_section_image(section_image_id,
                                     downsample=4,
                                     view='expression',
                                     rgb=[0.25,0.5,1])

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/70636013"
        "?downsample=4&rgb=0.25,0.5,1&view=expression",
        path)


def test_contrast_filter(image_api):
    '''
    '''
    path = '70636013.jpg'

    section_image_id = 70636013
    image_api.download_section_image(section_image_id,
                                     downsample=4,
                                     view='expression',
                                     contrast=[0.5,1])

    image_api.retrieve_file_over_http.assert_called_once_with(
        "http://api.brain-map.org/api/v2/section_image_download/70636013"
        "?downsample=4&contrast=0.5,1&view=expression",
        path)


def test_atlas_image_query(image_api):
    expected = "http://api.brain-map.org/api/v2/data/query.json?q=" + \
               "model::Atlas,rma::criteria,[id$eq1]," + \
               "rma::options[only$eqimage_type]," + \
               "pipe::list[type_name$is'image_type']," + \
               "model::AtlasImage,rma::criteria,[annotated$eqtrue]," + \
               "atlas_data_set(atlases[id$eq1])," + \
               "alternate_images[image_type$eq$type_name]," + \
               "rma::options[num_rows$eq'all']" + \
               "[order$eqsub_images.section_number]"

    adult_mouse_atlas_id = 1
    image_api.atlas_image_query(adult_mouse_atlas_id)

    image_api.json_msg_query.assert_called_once_with(expected)


def test_atlas_image_query_image_type_name(image_api):
    expected = "http://api.brain-map.org/api/v2/data/query.json?q=" + \
               "model::AtlasImage,rma::criteria,[annotated$eqtrue]," + \
               "atlas_data_set(atlases[id$eq1])," + \
               "alternate_images[image_type$eq'Atlas - Adult Mouse']," + \
               "rma::options[num_rows$eq'all']" + \
               "[order$eqsub_images.section_number]"

    adult_mouse_atlas_id = 1
    adult_mouse_image_type_name = 'Atlas - Adult Mouse'
    image_api.atlas_image_query(adult_mouse_atlas_id,
                                image_type_name=adult_mouse_image_type_name)

    image_api.json_msg_query.assert_called_once_with(expected)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
import scipy.sparse as sparse
from scipy.linalg import solve_banded
import logging


def get_diagonals_from_sparse(mat):
    ''' Returns a dictionary of diagonals keyed by offsets

    Parameters
    ----------
    mat: scipy.sparse matrix

    Returns
    -------
    dictionary: diagonals keyed by offsets
    '''

    mat_dia = mat.todia()  # make sure the matrix is in diagonal format

    offsets = mat_dia.offsets
    diagonals = mat_dia.data

    mat_dict = {}

    for i, o in enumerate(offsets):
        mat_dict[o] = diagonals[i]

    return mat_dict


def ab_from_diagonals(mat_dict):
    ''' Constructs value for scipy.linalg.solve_banded

    Parameters
    ----------
    mat_dict: dictionary of diagonals keyed by offsets

    Returns
    -------
    ab: value for scipy.linalg.solve_banded
    '''
    offsets = list(mat_dict.keys())
    l = -np.min(offsets)
    u = np.max(offsets)

    T = mat_dict[offsets[0]].shape[0]

    ab = np.zeros([l + u + 1, T])

    for o in offsets:
        index = u - o
        ab[index] = mat_dict[o]

    return ab


def error_calc(F_M, F_N, F_C, r):

    er = np.sqrt(np.mean(np.square(F_C - (F_M - r * F_N)))) / np.mean(F_M)

    return er


def error_calc_outlier(F_M, F_N, F_C, r):

    std_F_M = np.std(F_M)
    mean_F_M = np.mean(F_M)
    ind_outlier = np.where(F_M > mean_F_M + 2. * std_F_M)

    er = np.sqrt(np.mean(np.square(
        F_C[ind_outlier] - (F_M[ind_outlier] - r * F_N[ind_outlier])))) / np.mean(F_M[ind_outlier])

    return er


def ab_from_T(T, lam, dt):
    # using csr because multiplication is fast
    Ls = -sparse.eye(T - 1, T, format='csr') + \
        sparse.eye(T - 1, T, 1, format='csr')
    Ls /= dt
    Ls2 = Ls.T.dot(Ls)

    M = sparse.eye(T) + lam * Ls2
    mat_dict = get_diagonals_from_sparse(M)
    ab = ab_from_diagonals(mat_dict)

    return ab


def normalize_F(F_M, F_N):
    F_N_min, F_N_max = float(np.amin(F_N)), float(np.amax(F_N))

    # rescale so F_N is [0,1]
    F_M_s = (F_M - F_N_min) / (F_N_max - F_N_min)
    F_N_s = (F_N - F_N_min) / (F_N_max - F_N_min)

    return F_M_s, F_N_s


def alpha_filter(A=1.0, alpha=0.05, beta=0.25, T=100):
    return A * np.exp(-alpha * np.arange(T)) - np.exp(-beta * np.arange(T))


def validate_with_synthetic_F(T, N):
    """ Compute N synthetic traces of length T with known values of r, then estimate r.
    TODO: docs
    """
    af1 = alpha_filter()
    af2 = alpha_filter(alpha=0.1, beta=0.5)

    r_truth_vals = []
    r_est_vals = []

    for n in range(N):
        F_M_truth, F_N_truth, F_C_truth, r_truth = synthesize_F(T, af1, af2)
        results = estimate_contamination_ratios(F_M_truth, F_N_truth)

        r_est = results['r']

        r_truth_vals.append(r_truth)
        r_est_vals.append(r_est)

    return r_truth_vals, r_est_vals


def synthesize_F(T, af1, af2, p1=0.05, p2=0.1):
    """ Build a synthetic F_C, F_M, F_N, and r of length T
    TODO: docs
    """
    x1 = np.random.random(T) < p1
    F_C = np.convolve(af1, x1, mode='full')[:T]

    x2 = np.random.random(T) < p2
    F_N = np.convolve(af2, x2, mode='full')[:T]

    r = 2.0 * np.random.random()

    F_M = F_C + r * F_N

    return F_M, F_N, F_C, r


class NeuropilSubtract(object):
    """ TODO: docs
    """

    def __init__(self, lam=0.05, dt=1.0, folds=4):
        self.lam = lam
        self.dt = dt
        self.folds = folds

        self.T = None
        self.T_f = None
        self.ab = None

        self.F_M = None
        self.F_N = None

        self.r_vals = None
        self.error_vals = None
        self.r = None
        self.error = None

    def set_F(self, F_M, F_N):
        """ Break the F_M and F_N traces into the number of folds specified
        in the class constructor and normalize each fold of F_M and R_N relative to F_N.
        """

        F_M_len = len(F_M)
        F_N_len = len(F_N)

        if F_M_len != F_N_len:
            raise Exception(
                "F_M and F_N must have the same length (%d vs %d)" % (F_M_len, F_N_len))

        if self.T != F_M_len:
            logging.debug("updating ab matrix for new T=%d", F_M_len)
            self.T = F_M_len
            self.T_f = int(self.T / self.folds)
            self.ab = ab_from_T(self.T_f, self.lam, self.dt)

        self.F_M = []
        self.F_N = []

        for fi in range(self.folds):
            # F_M_i_s, F_N_i_s = normalize_F(F_M[fi*self.T_f:(fi+1)*self.T_f],
            #                               F_N[fi*self.T_f:(fi+1)*self.T_f])
            self.F_M.append(F_M[fi * self.T_f:(fi + 1) * self.T_f])
            self.F_N.append(F_N[fi * self.T_f:(fi + 1) * self.T_f])

    def fit_block_coordinate_desc(self, r_init=5.0, min_delta_r=0.00000001):
        F_M = np.concatenate(self.F_M)
        F_N = np.concatenate(self.F_N)

        r_vals = []
        error_vals = []
        r = r_init

        delta_r = None
        it = 0

        ab = ab_from_T(self.T, self.lam, self.dt)
        while delta_r is None or delta_r > min_delta_r:
            F_C = solve_banded((1, 1), ab, F_M - r * F_N)
            new_r = - np.sum((F_C - F_M) * F_N) / np.sum(np.square(F_N))
            error = self.estimate_error(new_r)

            error_vals.append(error)
            r_vals.append(new_r)

            if r is not None:
                delta_r = np.abs(r - new_r) / r

            r = new_r
            it += 1

        self.r_vals = r_vals
        self.error_vals = error_vals
        self.r = r_vals[-1]
        self.error = error_vals.min()

    def fit(self, r_range=[0.0, 2.0], iterations=3, dr=0.1, dr_factor=0.1):
        """ Estimate error values for a range of r values.  Identify a new r range
        around the minimum error values and repeat multiple times.
        TODO: docs
        """
        global_min_error = None
        global_min_r = None

        r_vals = []
        error_vals = []

        it_range = r_range
        it = 0

        it_dr = dr
        while it < iterations:
            it_errors = []

            # build a set of r values evenly distributed in a current range
            rs = np.arange(it_range[0], it_range[1], it_dr)

            # estimate error for each r
            for r in rs:
                error = self.estimate_error(r)
                it_errors.append(error)

                r_vals.append(r)
                error_vals.append(error)

            # find the minimum in this range and update the global minimum
            min_i = np.argmin(it_errors)
            min_error = it_errors[min_i]

            if global_min_error is None or min_error < global_min_error:
                global_min_error = min_error
                global_min_r = rs[min_i]

            logging.debug("iteration %d, r=%0.4f, e=%.6e",
                          it, global_min_r, global_min_error)

            # if the minimum error is on the upper boundary,
            # extend the boundary and redo this iteration
            if min_i == len(it_errors) - 1:
                logging.debug(
                    "minimum error found on upper r bound, extending range")
                it_range = [rs[-1], rs[-1] + (rs[-1] - rs[0])]
            else:
                # error is somewhere on either side of the minimum error index
                it_range = [rs[max(min_i - 1, 0)],
                            rs[min(min_i + 1, len(rs) - 1)]]
                it_dr *= dr_factor
                it += 1

        self.r_vals = r_vals
        self.error_vals = error_vals
        self.r = global_min_r
        self.error = global_min_error

    def estimate_error(self, r):
        """ Estimate error values for a given r for each fold and return the mean. """

        errors = np.zeros(self.folds)
        for fi in range(self.folds):
            F_M = self.F_M[fi]
            F_N = self.F_N[fi]
            F_C = solve_banded((1, 1), self.ab, F_M - r * F_N)
            errors[fi] = abs(error_calc(F_M, F_N, F_C, r))

        return np.mean(errors)


def estimate_contamination_ratios(F_M, F_N,
                                  lam=0.05, folds=4, iterations=3,
                                  r_range=[0.0, 2.0], dr=0.1, dr_factor=0.1):
    ''' Calculates neuropil contamination of ROI

    Parameters
    ----------
       F_M: ROI trace
       F_N: Neuropil trace

    Returns
    -------
    dictionary: key-value pairs
        * 'r': the contamination ratio -- corrected trace = M - r*N
        * 'err': RMS error
        * 'min_error': minimum error
        * 'bounds_error': boolean. True if error or R are outside tolerance
    '''

    ns = NeuropilSubtract(lam=lam, folds=folds)

    ns.set_F(F_M, F_N)

    ns.fit(r_range=r_range,
           iterations=iterations,
           dr=dr,
           dr_factor=dr_factor)

    # ns.fit_block_coordinate_desc()

    if ns.r < 0:
        logging.warning("r is negative (%f). return 0.0.", ns.r)
        ns.r = 0

    return {
        "r": ns.r,
        "r_vals": ns.r_vals,
        "err": ns.error,
        "err_vals": ns.error_vals,
        "min_error": ns.error,
        "it": len(ns.r_vals)
    }



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import patch, mock_open
from allensdk.config.model.description_parser import DescriptionParser
try:
    import __builtin__ as builtins
except:
    import builtins


@pytest.fixture
def pyconfig():
    file_1 = ("{\n"
              "    \"section_A\": [\n"
              "       {\n"
              "           \"prop_a\": \"val_a\",\n"
              "           \"prop_b\": \"val_b\"\n"
              "       },\n"
              "       {\n"
              "           \"prop_c\": \"val_c\",\n"
              "           \"prop_d\": \"val_d\"\n"
              "       }\n"
              "    ],\n"
              "    \"section_B\": [\n"
              "       {\n"
              "           \"prop_e\": \"val_e\",\n"
              "           \"prop_f\": \"val_f\"\n"
              "       },\n"
              "       {\n"
              "           \"prop_g\": \"val_g\",\n"
              "           \"prop_h\": \"val_h\"\n"
              "       }\n"
              "    ]\n"
              "}\n"
              )
    file_2 = ("{\n"
              "    \"section_B\": [\n"
              "       {\n"
              "           \"prop_i\": \"val_i\",\n"
              "           \"prop_j\": \"val_j\"\n"
              "       }\n"
              "    ],\n"
              "    \"section_C\": [\n"
              "       {\n"
              "           \"prop_k\": \"val_k\",\n"
              "           \"prop_l\": \"val_l\"\n"
              "       }\n"
              "    ]\n"
              "}\n"
              )

    with patch(builtins.__name__ + ".open",
               mock_open(
                   read_data=file_1)):
        parser = DescriptionParser()
        description = parser.read("mock_1.pycfg")

    with patch(builtins.__name__ + ".open",
               mock_open(
                   read_data=file_2)):
        parser = DescriptionParser()
        parser.read("mock_2.pycfg",
                    description)

    return description


def testAllSectionsPresent(pyconfig):
    assert('section_A' in pyconfig.data and
           'section_B' in pyconfig.data and
           'section_C' in pyconfig.data)
    assert(len(pyconfig.data.keys()) == 3)


def testSectionA(pyconfig):
    assert len(pyconfig.data['section_A']) == 2
    assert pyconfig.data['section_A'][0] == {
        'prop_a': 'val_a',
        'prop_b': 'val_b'}
    assert pyconfig.data['section_A'][1] == {
        'prop_c': 'val_c',
        'prop_d': 'val_d'}


def testSectionB(pyconfig):
    assert len(pyconfig.data['section_B']) == 3
    assert pyconfig.data['section_B'][0] == {
        'prop_e': 'val_e',
        'prop_f': 'val_f'}
    assert pyconfig.data['section_B'][1] == {
        'prop_g': 'val_g',
        'prop_h': 'val_h'}
    assert pyconfig.data['section_B'][2] == {
        'prop_i': 'val_i',
        'prop_j': 'val_j'}


def testSectionC(pyconfig):
    assert len(pyconfig.data['section_C']) == 1
    assert pyconfig.data['section_C'][0] == {
        'prop_k': 'val_k',
        'prop_l': 'val_l'}



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import logging


class HocUtils(object):
    '''A helper class for containing references to NEUORN.

    Attributes
    ----------
    h : object
        The NEURON hoc object.
    nrn : object
        The NEURON python object.
    neuron : module
        The NEURON module.
    '''
    _log = logging.getLogger(__name__)
    h = None
    nrn = None
    neuron = None

    def __init__(self, description):
        import neuron
        import nrn

        self.h = neuron.h
        HocUtils.neuron = neuron
        HocUtils.nrn = nrn
        HocUtils.h = self.h

        self.description = description
        self.manifest = description.manifest

        self.hoc_files = description.data['neuron'][0]['hoc']

        self.initialize_hoc()

    def initialize_hoc(self):
        '''Basic setup for NEURON.'''
        h = self.h
        params = self.description.data['conditions'][0]

        for hoc_file in self.hoc_files:
            HocUtils._log.info("loading hoc file %s" % (hoc_file))
            HocUtils.h.load_file(str(hoc_file))

        h('starttime = startsw()')

        if 'celsius' in params:
            h.celsius = params['celsius']

        if 'v_init' in params:
            h.v_init = params['v_init']

        if 'dt' in params:
            h.dt = params['dt']
            h.steps_per_ms = 1.0 / h.dt

        if 'tstop' in params:
            h.tstop = params['tstop']
            h.runStopAt = h.tstop
